using namespace std;
#include<iostream>
#include<cmath>

#include"InverseCompton.h"

/////////////////////////////////////////////////////////
void InverseCompton:: init(int debug_)
{

  cout<<">> InverseCompton:: init()"<<endl;

#include"Constants.h"



  constant=pi*re*re*me*me;
  initialized=1;
  debug=debug_;

  cout<<"<< InverseCompton:: init()"<<endl;
}
/////////////////////////////////////////////////////////

double InverseCompton::IC_sigma_KN(double Ee, double Eph, double eta, double Egamma)
{
  // Inverse Compton differential cross-section using Klein-Nishina cross-section.

  // units: cm^2 MeV^-1
  // uses equations (6)+(7) of    Orlando & Strong A&A 480, 847 (2008)
  // and  equation  (9)     of Moskalenko & Strong ApJ 528, 357 (2000)
  // stand-alone form, all constants defined here
  // energies in MeV
  // Ee     = electron energy
  // Eph    = target photon energy
  // eta    = angle between electron and target photon in radians
  // Egamma = gamma ray energy

#include"Constants.h"

  double result;
  if(Egamma>=Ee) {result=0.; return result;}
  if(initialized!=1) init(0); // just in case


  double g    = Ee/me;                      // electron Lorentz factor
  double beta = sqrt(1.0-1.0/(g*g));        // electron velocity/c
  double v    = Egamma/Ee;                  // fractional energy transfer
  double Ephd = g*Eph*(1.0+beta*cos(eta));  // energy of target photon in electron frame

  double Egamma_max = 2.0*g*Ephd/(1.0+2.0*Ephd/me); //NB MS2000 eq (9) Ephd has units of me
  if(Egamma>=Egamma_max) {result=0.; return result;}

  
 
  result= constant/(Eph*Ee*Ee)
         *(
           (me/Ephd)*(me/Ephd)*v*v/((1.0-v)*(1.0-v))
                -2.0*(me/Ephd)  *v/ (1.0-v)
                                   +(1.0-v) + 1.0/(1.0-v)
          );
  

 
  if(debug==1)
  cout<<"IC_sigma_KN: Ee="<<Ee<<" Eph="<<Eph<<" eta="<<eta<<" Egamma="<<Egamma
      <<" g="<<g<<" beta="<<beta
      <<" Ephd="<< Ephd <<" me/Ephd="<< me/Ephd<<" v="<<v<<" Egamma_max="<<Egamma_max
      << " result="<<result<<endl;


  return result;
};


///////////////////////////////////////////////////////////////////////////

  void  InverseCompton:: IC_emiss(valarray<double> Ee,   valarray<double> Eph, double eta,valarray<double> Egamma,
                                  valarray<double> fluxe,valarray<double> nph,            valarray<double> &emiss,
                                  int method )

{

  if(debug==1) cout<<">> InverseCompton::IC_emiss()"<< " method="<<method<<endl;

  // assuming both electron and target photon energy grids are logarithmic

  ///  emiss.resize(Egamma.size());

  emiss=0.0;

 if(method==1)
 {

  double IC_sigmaKN;

  for (int iEe    =0;iEe    <Ee    .size();iEe    ++)
  for (int iEph   =0;iEph   <Eph   .size();iEph   ++)
  for (int iEgamma=0;iEgamma<Egamma.size();iEgamma++)
  {
    IC_sigmaKN      = IC_sigma_KN   (Ee[iEe],  Eph[iEph], eta, Egamma[iEgamma]);

    emiss[iEgamma] += IC_sigmaKN* fluxe[iEe] * nph[iEph]
                                    *Ee[iEe] * Eph[iEph];

    if(debug==1) cout<<"IC_emiss: method="<<method<<" Egamma="<< Egamma[iEgamma]<<" Ee="<<Ee[iEe]<<" Eph ="<<Eph[iEph]
                <<"  IC_sigmaKN= "<<IC_sigmaKN<<" emiss="<<emiss[iEgamma]<<endl;
  }


  emiss *= log(Ee[1]/Ee[0])*log(Eph[1]/Eph[0]);// NB assumes at least 2 values in spectra  !

 }// method==1

 ////////////// optimized inline method

 if(method==2)
 {
#include"Constants.h"
 if(initialized!=1) init(0); // just in case

  double IC_sigmaKN;

  for (int iEe    =0;iEe    <Ee    .size();iEe    ++)
  {
  double g    = Ee[iEe]/me;                      // electron Lorentz factor
  double beta = sqrt(1.0-1.0/(g*g));        // electron velocity/c

  for (int iEph   =0;iEph   <Eph   .size();iEph   ++)
  {
 
   double Ephd       = g*Eph[iEph]*(1.0+beta*cos(eta));  // energy of target photon in electron frame
   double Egamma_max = 2.0*g*Ephd/(1.0+2.0*Ephd/me);

   double term1=  constant/(Eph[iEph]*Ee[iEe]   *Ee[iEe])
                                 * fluxe[iEe] * nph[iEph]
                                     *Ee[iEe] * Eph[iEph];

  for (int iEgamma=0;iEgamma<Egamma.size();iEgamma++)
  {
   

  if(Egamma[iEgamma] < Ee[iEe]) 
  {
    
   if(Egamma[iEgamma]< Egamma_max) 
   {

  
    double v    = Egamma[iEgamma]/Ee[iEe];                  // fractional energy transfer
    double v1   = 1.0-v;
    double v2   = v/v1;
    double v3   = v1 + 1.0/v1;

    double me_Ephd_v2= me/Ephd*v2;

 

    double term2= me_Ephd_v2*(me_Ephd_v2 -2.0) + v3;


    emiss[iEgamma] +=   term1 * term2;

     /* for reference to follow the formula above

       (me/Ephd)*(me/Ephd)*v*v/((1.0-v)*(1.0-v))
            -2.0*(me/Ephd)  *v/ (1.0-v)
                               +(1.0-v) + 1.0/(1.0-v);
    
    */

  if(debug==1)
  cout<<"IC_emiss : method="<<method
      <<" Ee="<<Ee[iEe]<<" Eph="<<Eph[iEph]<<" eta="<<eta<<" Egamma="<<Egamma[iEgamma]
      <<" g="<<g<<" beta="<<beta
      <<" Ephd="<< Ephd <<" me/Ephd="<< me/Ephd<<" v="<<v<<" Egamma_max="<<Egamma_max
      << " IC_sigmaKN="<<IC_sigmaKN<<endl;

 
     }// Egamma < Egamma_max
  }//   Egamma < Ee

    if(debug==1) cout<<"IC_emiss : method="<<method<<" Egamma="<< Egamma[iEgamma]<<" Ee="<<Ee[iEe]<<" Eph ="<<Eph[iEph]
                <<"  IC_sigmaKN= "<<IC_sigmaKN<<" emiss="<<emiss[iEgamma]<<endl;

 

    }//Egamma
   }//Eph
  }//Ee

  
  emiss *= log(Ee[1]/Ee[0])*log(Eph[1]/Eph[0]);// NB assumes at least 2 values in spectra  !

 }// method==2


  if(debug==1)  cout<<"<< InverseCompton::IC_emiss()"<<endl;
  return;
};


//////////////////////////////
// other functions foreseen for arrays of energies at constant eta, etc, for more efficient computation
//////////////////////////////






/////////////////////////////////////////////////////////
void InverseCompton::test()
{

  cout<<">> InverseCompton::test()"<<endl;
  cout<<   " testing IC_sigma_KN"<<endl;

  int debug_=1;
  init(debug_);

  double Ee,  Eph, eta,  Egamma;
  Ee=1e3;
  Eph=1e-3*1.e-6;// NB MeV 
  Egamma=1e2;
  eta=0.5;

  double pi=acos(-1.0);
  double factor=10;
  double deta  =0.01;

  for(Ee    =1;   Ee    <1e6;    Ee *=factor)
  for(Eph   =1e-4;Eph   <1e1;    Eph*=factor)
  for(Egamma=0.1; Egamma<1e6;Egamma *=factor)
  for(eta   =0;   eta   < pi;    eta+=deta  )
  {
  double IC_sigmaKN=IC_sigma_KN( Ee,  Eph, eta,  Egamma);

  if(IC_sigmaKN>0.)
  cout<<"Ee="<<Ee<<" Eph="<<Eph<<" Egamma="<<Egamma<<" eta="<<eta
      << " IC_sigma_KN="<<IC_sigmaKN<<endl;
  }



//----------------------------------- test emissivity

  cout<<" testing emissivity"<<endl;

  int nEe    =400;   // NB must be >1 for correct run
  int nEph   =200   ; // NB must be >1 for correct run
  int nEgamma=100 ;

  double  Eemin       =1.0;
  double  Eefactor    =1.05;
  double  Ephmin      =1.0e-4 * 1e-6;//NB MeV
  double  Ephfactor   =1.05;
  double  Egammamin   =1.0;
  double  Egammafactor=pow(10.,0.1);

  double ge=3.0; // electron spectral index
 
  valarray<double> Ee_grid     (nEe);
  valarray<double> Eph_grid    (nEph);
  valarray<double> Egamma_grid (nEgamma);
  valarray<double> fluxe       (nEe);
  valarray<double> nph         (nEph);
  valarray<double> emiss       (nEgamma);

  int method = 2;
  debug=0;
  

  for (int iEgamma=0;iEgamma<Egamma_grid.size();iEgamma++) Egamma_grid [iEgamma]=Egammamin*pow(Egammafactor,iEgamma);
 
  for (int iEph   =0;iEph   <Eph_grid   .size();iEph   ++) Eph_grid    [iEph]   =Ephmin   *pow(Ephfactor,   iEph   );
  for (int iEph   =0;iEph   <Eph_grid   .size();iEph   ++) nph         [iEph]   = 1.0e6;// per MeV

  for (int iEe    =0;iEe    <Ee_grid    .size();iEe    ++) Ee_grid     [iEe]    =Eemin    *pow(Eefactor,    iEe    );
  for (int iEe    =0;iEe    <Ee_grid    .size();iEe    ++) fluxe       [iEe]    =          pow(Ee_grid[iEe], -ge   );

  // Ackermann etal Fermi-LAT arXiv:1109.052 electrons fit, converting from GeV m^2 to MeV cm^2 
  ge=3.19;
  for (int iEe    =0;iEe    <Ee_grid    .size();iEe    ++) fluxe       [iEe]    = 2.07e-9* pow(Ee_grid[iEe]/2.e4, -ge   );
   for (int iEe    =0;iEe    <Ee_grid    .size();iEe++)     cout<<"Ee= "    << Ee_grid    [iEe]<<" fluxe="<<fluxe[iEe]<<endl;
   for (int iEph   =0;iEph   <Eph_grid   .size();iEph   ++) cout<<"Eph= "   << Eph_grid   [iEph]<<" nph=" <<  nph[iEph]<<endl;

  eta=0.1; // 0= head on

  for(method=1;method<=2;method++)
  {
   IC_emiss(Ee_grid, Eph_grid, eta, Egamma_grid,
	   fluxe,   nph,           emiss,
           method );

   

   for (int iEgamma=0;iEgamma<Egamma_grid.size();iEgamma++)
       cout<<" method="<<method<<" Egamma= "<< Egamma_grid[iEgamma]<<" emiss="<<emiss[iEgamma]<<endl;

  }//method
  

  cout<<"<< InverseCompton::test()"<<endl;
  return;
}

///////////////////////////// Test program
/*
int main()
{

  InverseCompton inversecompton;



  inversecompton.test();

  return 0;
}
*/
